import numpy as np
from time import sleep

from epics import caget,caput,PV

from iexcode.instruments.IEX_endstations import Endstation
from iexcode.instruments.staff import staff_detector_dictionary

from iexcode.instruments.files_and_folders import check_run,make_user_folders,folder_mda
from iexcode.instruments.staff import staff_detector_dictionary
from iexcode.instruments.logfile import logfile_name_set,logfile_header

from iexcode.instruments.conversions_constants import *
from iexcode.instruments.utilities import *

from iexcode.instruments.userCalcs import userStringSeq_clear, userStringSeq_pvs
from iexcode.instruments.scanRecord import ScanRecord
from iexcode.instruments.xrays import xrays_detector_dictionary,xrays_reset,xrays_get_all

from iexcode.instruments.Motors import Motors
from iexcode.instruments.current_amplifiers import *
from iexcode.instruments.gate_valves import valve_close, branch_valves
from iexcode.instruments.shutters import branch_shutter_close
from iexcode.instruments.slits import slit3C_get

from iexcode.instruments.Lakeshore_335 import Lakeshore_reset
from iexcode.instruments.electron_analyzer import EA_init_Motors, folders_EA,EA

global BL
global tey,ca15
global ARPES_Motors

#############################################################################
def ARPES_init(set_folders=True,reset=True,**kwargs):
    """
        kwargs:
        set_folders: sets the mda and EA folders; default => False
        xrays: sets global variable; default => True   
    """
    kwargs.setdefault('scan_ioc','29idARPES:')
    kwargs.setdefault('xrays',True)
    kwargs.setdefault('BL_mode','user')

    #scan
    if kwargs['BL_mode']=='staff':
        detector_dictionary = staff_detector_dictionary()
    else:
        detector_dictionary = ARPES_detector_dictionary()
    mda_scanRecord = ScanRecord(kwargs['scan_ioc'],detector_dictionary,
    ARPES_trigger_dictionary(),ARPES_scan_before_sequence(),ARPES_scan_after_sequence())

    #endstation
    BL=Endstation('ARPES',kwargs['scan_ioc'],kwargs['xrays'],kwargs['BL_mode'],mda_scanRecord,
    ARPES_log_header,ARPES_log_entries,ARPES_motor_dictionary)

    #EA
    try: 
        EA.get()
        EA_init_Motors(ARPES_Motors)
    except:
        print("\n\n NOTE: Scienta IOC is not running - start IOC and %run Macros_29id/ScanFunctions_EA.py\n\n")
    
    #global detectors
    tey = Keithley('c',1)
    ca15 = Keithley('b',15)

    #setting folders
    if 'set_folders':
        if BL.mode == 'staff':
            user_name = 'staff'
    else:
        user_name = input('user name: ')
    folders_ARPES(user_name,**kwargs)
    
    #resetting
    if 'reset':
        ARPES_reset()

    #motors
    physical_motors = ['x','y','z','th','chi','phi']
    psuedo_motors = ['focus']
    ARPES_Motors = Motors('ARPES',ARPES_motor_dictionary(),physical_motors,psuedo_motors)
    
    print ('ARPES_init')
##############################################################################################################
##############################             ARPES detectors and motors         ##############################
##############################################################################################################
def ARPES_detector_list():
    """
    list of detectors to trigger 

    Previously: part of Detector_List
    """
    ca_list=[["c",1],["b",15],["b",4],["b",13]]
    return ca_list

def ARPES_detector_dictionary(xrays=True):
    """
    returns a dictionary of the default detectors for the scan record

    Previously: Detector_Default
    """
    det_dict={
        14:"29idc:ca2:read",
        16:"29idc:ca1:read",
        17:EA._statsPlugin+"Total_RBV",
        18:"29idARPES:LS335:TC1:IN1",
        19:"29idARPES:LS335:TC1:IN2",
        }
    if xrays: 
        det_dict.update(xrays_detector_dictionary())
    return det_dict

def ARPES_motor_dictionary():
    """
    returns a dictionary with  {name:[rbv,val,spmg,pv]}
    """
    
    motor_nums = {'x': 1,
      'y':2,
      'z':3,
      'th':4,
      'chi':5,
      'phi':6,
      }
    
    motor_dictionary={}
    for name in motor_nums.keys():
        pv = '29idc:m'+str(motor_nums[name])
        motor_dictionary.update({name:[pv+'.RBV',pv+'.VAL',pv+'.SPMG',pv]})

    return motor_dictionary

def ARPES_extra_pvs():
    """
    used to get the PV associated with a given pnuemonic

    """
    d={
        "SES_slit":"29idc:m8.RBV",
        "TA":"29idARPES:LS335:TC1:IN1", 
        "TB":"29idARPES:LS335:TC1:IN2",
        "tey1":"29idc:ca1:read",
        "tey2":"29idc:ca2:read",
    }
    return d
    
##############################################################################################################
##############################                 setting folder                   ##############################
##############################################################################################################
def folders_ARPES(user_name,**kwargs):
    """
    Create and sets (if create_only=False) all the folders the current run and ARPES user 
    Sets the FileName for logging to be "UserName/YYYYMMDD_log.txt using logname_Set()"
    
    **kwargs:
        set_folders = True (default); set the mda and EA scanRecords
                    = False; only makes the folders (was create_only)
        run: run cycle e.g. 2022_1 (default => check_run() )
        ftp = True / False (default); print what needs to be added to the ftps crontab and create additional ftp folders
        mda_ioc: will overwrite the default ioc for mda scans
        debug
           
    """
    kwargs.setdefault('set_folders',True)
    kwargs.setdefault('run',check_run())
    kwargs.setdefault('ftp',False)
    kwargs.setdefault('debug',False)

    run = kwargs['run']

    if kwargs['debug']:
        print("run,folder,user_name,ioc,ftp: ",run,BL.folder,user_name,BL.ioc,kwargs['ftp'])

    # Create User Folder:
    make_user_folders(run,BL.folder,user_name,BL.endstation,ftp=kwargs['ftp'])
    sleep(5)

    if kwargs["set_folders"]:
        # Set up MDA folder:
        folder_mda(run,BL.folder,user_name,BL.prefix,BL.ioc)
        logfile_name_set(BL.endstation)
        logfile_header(BL.endstation,BL.ioc,ARPES_log_header())

        
        #Set up Scienta folders:
        try:
            userPath = "/net/s29data/export/data_29id"+BL.folder+"/"+run+"/"+user_name+"/"
            folders_EA(userPath,filePrefix="EA")
        except:
            print_warning_message("EA ioc is not running, cannot set folder")

def ARPES_reset():
    """
    resets scanRecord, current amplifiers, mono limits and lakeshore
    """
    #resetting the scanRecord
    BL.mda.reset()

    #resetting the current amplifiers
    if BL.xray:
        ca_reset_all()
    else:
        tey.reset()

    #resetting mono and anyother beamline stuff
    if BL.xrays:
        xrays_reset()

    #reseting the ARPES Lakeshore
    pv = "29idARPES:LS335:"
    LS_355_defaults = {
        "TC1:read.SCAN":".5 second",
        "TC1:OUT1:Cntrl":"A",
        "TC1:OUT2:Cntrl":"B",
        "TC1:OUT1:Mode":"Closed Loop"
    }
    Lakeshore_reset(pv,LS_355_defaults)

##############################################################################################################
##############################                    get all                 ##############################
##############################################################################################################
def ARPES_get_all(verbose=True):
    """
    returns a dictionary with the current status of the ARPES endstation and exit slit
    """
    vals={}

    #sample position
    motor_dictionary = ARPES_motor_dictionary()
    for motor in motor_dictionary.keys():
        vals[motor]=ARPES_Motors.get(motor,verbose=False)
        
    #endstation pvs
    extra_pvs = ARPES_extra_pvs()
    for key in extra_pvs.keys():
        vals.update(key,caget(extra_pvs[key]))
    vals.update('exit_slit',slit3C_get())

    #beamline info
    if BL.xray:
        beamline_info = xrays_get_all()
        vals.update(beamline_info) 

    if verbose:
        for key in vals:
            print(key+" = "+vals[key])
    return vals



##############################################################################################################
##############################             logging           ##############################
##############################################################################################################
def ARPES_log_header():
    header_list={
        'EA': "scan,x,y,z,th,chi,phi,T,scan_mode,E1,E2,step,i,f,PE,lens_mode,SES slit #,hv,exit_slit,GRT,ID_SP,ID_RBV,ID_Mode,ID_QP,TEY1,TEY2,time,comment",
        'ARPES':"scan,motor,start,stop,step,x,y,z,th,chi,phi,T,hv,exit_slit,GRT,ID_SP,ID_RBV,ID_Mode,ID_QP,TEY_1,TEY_2,time,comment"
        }
    return header_list

def ARPES_log_entries():
    """
    endstation info for log file

    Previously: scanlog
    """
    vals = ARPES_get_all(verbose=False)
    x = vals['x']
    y = vals['y']
    z = vals['z']
    th = vals['th']
    chi = chi['chi']
    phi = vals['phi']
    TA = vals['TA']
    TB = vals['TB']
    tey1 = vals['tey1']
    tey2 = vals['tey2']

    entry_list = ["x","y","z","th","chi","phi","TA","TB","tey1","tey2"]
    pv_list = [ x, y, z, th, chi, phi, TA, TB, tey1,tey2]  
    format_list = [".2f",".2f",".2f",".2f",".2f",".2f",".2f",".2f","1.2e","1.2e"]

    return entry_list,pv_list,format_list

##############################################################################################################
##############################             ARPES scanRecord           ##############################
##############################################################################################################     
def ARPES_scan_before_sequence(**kwargs):
    """
    writes the user string sequence to happen at the beginning of a scan
    returns before_scan_pv = pv for userStringSeq for before scan

    **kwargs
        seq_num: userStringSeq number in ioc => 9 (default)

    Previously: BeforeScan_StrSeq
    """
    kwargs.setdefault(seq_num,9)
    seq_num=kwargs['seq_num']

    before_scan_pv,before_scan_proc = userStringSeq_pvs(BL.ioc, seq_num)

    #clear and write the before scan user sequence
    userStringSeq_clear(BL.ioc,seq_num)
    caput(before_scan_pv+".DESC","Before Scan")

    #sequence put CAs in passive
    ca_list = ARPES_detector_list()
    for (i,ca) in enumerate(ca_list):
        ca_pv = Keithley_pv(ca[0], ca[1])+':read.SCAN PP NMS'
        caput(before_scan_pv+".LNK" +str(i+1),ca_pv)
        caput(before_scan_pv+".STR" +str(i+1),"Passive")

    return before_scan_proc

def ARPES_ca_live_sequence(**kwargs):
    """
    """
    kwargs.setdefault('seq_num',7)
    ca_live_sequence_proc = ca_live_sequence(BL.ioc,kwargs['seq_num'],ARPES_detector_list())
    return ca_live_sequence_proc

def ARPES_scan_after_sequence(scan_dim, **kwargs):
    """
    writes the user string sequence to happen at the end of a scan
    returns after_scan_pv = pv for userStringSeq for after scan

    **kwargs
        seq_num: userStringSeq number in ioc => 10 (default)
        snake: for snake scanning => False (default)
    Previously: AfterScan_StrSeq
    """
    kwargs.setdefault(seq_num,10)
    kwargs.setdefault('snake',False)
    seq_num=kwargs['seq_num']

    after_scan_pv,after_scan_proc = userStringSeq_pvs(BL.ioc, seq_num)
    
    #clear and write the after scan user sequence
    userStringSeq_clear(BL.ioc,seq_num)
    caput(after_scan_pv+".DESC","After Scan")
        
    ## Put All relevant CA back in live mode
    ca_live_sequence_proc = ARPES_ca_live_sequence(**kwargs)
    caput(after_scan_pv+".LNK1",ca_live_sequence_proc+" PP NMS")
    caput(after_scan_pv+".DO1",1)

    scan_pv = BL.ioc+"scan"+str(scan_dim)
    ## Put scan record back in absolute mode
    caput(after_scan_pv+".LNK2",scan_pv+".P1AR")
    caput(after_scan_pv+".STR2","ABSOLUTE")

    ## Put Positionner Settling time to 0.1s
    caput(after_scan_pv+".LNK3",scan_pv+".PDLY NPP NMS")
    caput(after_scan_pv+".DO3",0.1)

    ## Clear DetTriggers 2:
    caput(after_scan_pv+".LNK4",scan_pv+".T2PV NPP NMS")    #remove trigger 2 (used for EA) after scan

    if kwargs['snake']:
        snake_dim = scan_dim - 1
        #snake_proc= ARPES_snake_pv(snake_dim,enable=True)
        #caput(after_scan_pv+".LNK10",ARPES_snake_pv()+"PP NMS")
        #caput(after_scan_pv+".D10",1)

    return after_scan_proc


def ARPES_detector_triggers_sequence(**kwargs):    # do we need to add 29idb:ca5 ???
    """
    """
    kwargs.setdefault(seq_num,8)
    seq_num=kwargs['seq_num']

    detector_triggers_pv,detector_triggers_proc = userStringSeq_pvs(BL.ioc, seq_num)
       
    #clear the userStringSeq
    userStringSeq_clear(BL.ioc,seq_num=kwargs['seq_num'])
    caput(detector_triggers_pv+".DESC","ARPES_Trigger1")
    
    ca_list = ARPES_detector_list
    last = len(ca_list)
    for i,ca in enumerate(ca_list):
        ca_pv = Keithley_pv(ca[0], ca[1])+':read.PROC CA NMS'
        caput(detector_triggers_pv+".LNK" +str(i+1),ca_pv)
        caput(detector_triggers_pv+".WAIT"+str(i+1),"After"+str(last))

    return detector_triggers_proc

def ARPES_trigger_dictionary():
    """
    need to do something
    """
    trigger_dictionary = {
        1:ARPES_detector_triggers_sequence(),
    }
    return trigger_dictionary


##############################################################################################################
##############################             ARPES Motor Scan Set Up        ##############################
##############################################################################################################
def ARPES_motor_encoder_sync():
    ioc = "29idc:"
    for motor in [1,2,3,4]:
        pv = ioc+"m"+str(motor)+".SYNC"
        caput(pv,1)
        print("synch encoder: "+pv)




####################################################################################################
################    ARPES Positions (LEED,transfer, measure)    #####################################
#####################################################################################################
def _ARPES_DefaultPosition(destination):
    """
    Default ARPES positions in DIAL units
    """
    DefaultPosition={
    'measure':(0.0,0.0,-13.5,0.0,0.0,0.0),
    'LEED':(-0.0, 0.0, -141.5, 89.5, 0.0, 0.0),
    'transfer':(-0,0.0,-141.5,180.0,0.0,0.0),
    }
    if destination in DefaultPosition:
        pos=DefaultPosition[destination]
    else:
        pos=(None,None,None,None,None,None)
    return pos

def _ARPES_MoveSequence(destination):
    """
    Moves the ARPES manipulator x,y,z,th to a given destination (does not change chi or phi)
    Resets the limits to help protect from crashes 
        such as: phi/chi motor body into the strong back, manipulator into the LEED, manipulator into the wobble stick
        
    DO NOT MAKE BIG MOVE WITH OUT WATCHING THE MANIPULATOR!!!! 
        if the theta encoder stops reading the manipulator will CRASH!!!!
    """
    (x,y,z,th,chi,phi)=_ARPES_DefaultPosition(destination)
    motor_dictionary = ARPES_motor_dictionary()
    if x is None:
        print("Not a valid destination")
    else:
        #reset limits to default values
        _ARPES_limits_set(None)
        #print destination
        print(("Moving to "+destination+": "+str(round(x,1))+","+str(round(y,1))+","+str(round(z,1))+","+str(round(th,1)))+'     Time:'+time.strftime("%H:%M"))
        # move x and y to zero for all moves (helps protect the motors from crashing into the strong back)
        caput(motor_dictionary['x'][3]+".DVAL",0,wait=True,timeout=18000)
        caput(motor_dictionary['y'][3]+".DVAL",0,wait=True,timeout=18000)
        # moves z and th simultaneously and monitors the .DMOV
        caput(motor_dictionary['z'][3]+".DVAL",z,wait=False,timeout=18000)
        caput(motor_dictionary['th'][3]+".DVAL",th,wait=False,timeout=18000)
        while True:
            sleep(5)
            Test=caget(motor_dictionary['z'][3]+".DMOV")*caget(motor_dictionary['th'][3]+".DMOV")
            if(Test==1):
                break
        # move x and y to final position
        caput(motor_dictionary['x'][3]+".DVAL",x,wait=True,timeout=18000)
        caput(motor_dictionary['y'][3]+".DVAL",y,wait=True,timeout=18000)
        print(("Arrived at "+destination+": "+str(round(x,1))+","+str(round(y,1))+","+str(round(z,1))+","+str(round(th,1)))+'     Time:'+time.strftime("%H:%M"))
        #set the limits for this position
        _ARPES_limits_set(destination)

def _ARPES_limits_set(destination):
    """
    Resets the ARPES motor limits in Dial cooridinates(remember that z is negative)
        destination: 'measures' or 'LEED',
        else: sets to the full range
    ARPES_MoveSequence(destination) minimizes that chance of a crash and sets limits

    Previously: ARPES_LimitsSet
    """
    motor_dictionary = ARPES_motor_dictionary()
    if destination == 'measure':
        (x,y,z,th,chi,phi)=_ARPES_DefaultPosition("measure")
        limits={'x':[5.5,-7],'y':[7,-5],'z':[-6,-310],'th':[th+35,th-25],'chi':[45,-15],'phi':[120,-120]}
    elif destination == 'LEED':
        (x,y,z,th,chi,phi)=_ARPES_DefaultPosition("LEED")
        limits={'x':[5.5,-7],'y':[6,-1],'z':[z+5,z-5],'th':[th+2,th-2],'chi':[45,-15],'phi':[120,-120]}
    else:
        limits={'x':[5.5,-7],'y':[7,-5],'z':[-6,-310],'th':[240,-70],'chi':[45,-15],'phi':[120,-120]}
    #setting the limits for each motor in Dial
    for m in limits:
        caput(motor_dictionary[m][3]+'.DHLM',limits[m][0])
        caput(motor_dictionary[m][3]+'.DLLM',limits[m][1])
    if destination is not None:
        print("Limits have been reset for "+destination+" position")
    else:
        print("Limits have been reset for full range")
####################################################################################################
def ARPES_transfer(chi=0,phi=0,**kwargs):
    """
    Moves the ARPES manipulator to the default transfer position
    kwargs:
        EA_HV_Off=True; Turns off the EA HV
        Close_CBranch=True;  closes the C-shutter and the C-valve (main chamber to BL)
    """
    ARPESgo2("transfer",**kwargs)

def ARPES_measure(**kwargs):
    """
    Moves to ARPES motors x,y,z,th to the default measurement position
    kwargs
        chi=None # specify a value to move chi
        phi=None # specifiy a value to move phi
    """
    ARPESgo2("measure",**kwargs)

def ARPES_LEED(**kwargs):
    """
    Moves to ARPES motors x,y,z,th to the default LEED position
    kwargs:
        EA_HV_Off=True; Turns off the EA HV
        Close_CBranch=True;  closes the C-shutter and the C-valve (main chamber to BL)
        
        chi=None # specify a value to move chi
        phi=None # specifiy a value to move phi
    """
    ARPESgo2("LEED",**kwargs)    
####################################################################################################
def ARPESgo2(destination,**kwargs):
    """
    Moves the ARPES manipulator to the default position: 
    destination: "transfer", "measure", "LEED"

    kwargs:
        EA_off = True; turns off EA HV
        shutter_close = True; closes the C-shutter
        valve_close = True; closes the C-valve
        
        chi=None => doesn't move, otherwise specify a value to move chi
        phi=None => doesn't move, otherwise specifiy a value to move phi 

    """


    kwargs.setdefault("EA_off",True)
    kwargs.setdefault("shutter_close",True)
    kwargs.setdefault("valve_close",True)

    kwargs.setdefault("chi",None)
    kwargs.setdefault("phi",None)
    
    ARPES_safe_state(**kwargs)
    
    #Move x,y,z,th
    _ARPES_MoveSequence(destination)
    #Move chi and phi back to 0
    motor_dictionary = ARPES_motor_dictionary()
    if kwargs["chi"] is not None:
        caput(motor_dictionary['chi'][1],kwargs["chi"])
    if kwargs["phi"] is not None:
        caput(motor_dictionary['phi'][1],kwargs["phi"])


##############################################################################################################
##############################             ARPES pvs for GUIs (caQtDM)        ##############################
##############################################################################################################
def ARPES_StringCalc_trigger2_active():
    """
    setup a string calc to be used by caQtDM to indicate that the trigger 2 is filled
    """
    ioc="29idARPES:"
    pv=ioc+"userStringCalc2"
    caput(pv+".DESC","scan1 trigger2 calc")
    caput(pv+".INAA",ioc+"scan1.T2PV CP NMS")
    caput(pv+".CALC$","AA")
    
    caput(ioc+"userStringCalcEnable.VAL","Enable")



##############################################################################################################
##############################             ARPES safestate        ##############################
##############################################################################################################
def ARPES_safe_state(**kwargs):
    """
    puts the C-branch in a safe state, 
    **kwargs
        EA_off = True; turns off EA HV
        shutter_close = True; closes the C-shutter
        valve_close = True; closes the C-valve
    """
    kwargs.setdefault("EA_off",True)
    kwargs.setdefault("shutter_close",True)
    kwargs.setdefault("valve_close",True)
    
    if kwargs["EA_off"]:
        try:
            EA.off(quiet=False)
        except:
            print('EA is not running, visually confirm HV is off')

    if kwargs['valve_close']:
       valve_close(branch_valves('c'), verbose=True)

    if kwargs['shutter_close']:
        branch_shutter_close('c')

##############################################################################################################
##############################              ARPES motors       ##############################
##############################################################################################################
### x ###
def mvx(val,**kwargs):
    """ moves x motor in the endstation"""
    ARPES_Motors.move('x',val,**kwargs)

def dmvx(val):
    """ relative x motor in the endstation"""
    ARPES_Motors.move('x',val,relative=True)

def scanx(start,stop,step,**kwargs):
    """ scans the x motor in the endstation"""
    ARPES_Motors.scan("x",start,stop,step,**kwargs)

def dscanx(start,stop,step):
    """relative scans of the x motor in the endstation """
    ARPES_Motors.scan("x",start,stop,step,relative=True)

### y ###
def mvy(val,**kwargs):
    """ moves y motor in the endstation"""
    ARPES_Motors.move('y',val,**kwargs)

def dmvy(val):
    """ relative y motor in the endstation"""
    ARPES_Motors.move('y',val,relative=True)

def scany(start,stop,step,**kwargs):
    """ scans the y motor in the endstation"""
    ARPES_Motors.scan("y",start,stop,step,**kwargs)

def dscany(start,stop,step):
    """relative scans of the y motor in the endstation """
    ARPES_Motors.scan("y",start,stop,step,relative=True)

### z ###
def mvz(val,**kwargs):
    """ moves z motor in the endstation"""
    ARPES_Motors.move('z',val,**kwargs)

def dmvz(val):
    """ relative z motor in the endstation"""
    ARPES_Motors.move('z',val,relative=True)

def scanz(start,stop,step,**kwargs):
    """ scans the z motor in the endstation"""
    ARPES_Motors.scan("z",start,stop,step,**kwargs)

def dscanz(start,stop,step):
    """relative scans of the z motor in the endstation """
    ARPES_Motors.scan("z",start,stop,step,relative=True)

### th ###
def mvth(val,**kwargs):
    """ moves th motor in the endstation"""
    ARPES_Motors.move('th',val,**kwargs)

def dmvth(val):
    """ relative th motor in the endstation"""
    ARPES_Motors.move('th',val,relative=True)

def scanth(start,stop,step,**kwargs):
    """ scans the th motor in the endstation"""
    ARPES_Motors.scan("th",start,stop,step,**kwargs)

def dscanth(start,stop,step):
    """relative scans of the th motor in the endstation """
    ARPES_Motors.scan("th",start,stop,step,relative=True)

### chi ###
def mvchi(val,**kwargs):
    """ moves chi motor in the endstation"""
    ARPES_Motors.move('chi',val,**kwargs)

def dmvchi(val):
    """ relative chi motor in the endstation"""
    ARPES_Motors.move('chi',val,relative=True)

def scanchi(start,stop,step,**kwargs):
    """ scans the chi motor in the endstation"""
    ARPES_Motors.scan("chi",start,stop,step,**kwargs)

def dscanchi(start,stop,step):
    """relative scans of the chi motor in the endstation """
    ARPES_Motors.scan("chi",start,stop,step,relative=True)

### phi ###
def mvphi(val,**kwargs):
    """ moves phi motor in the endstation"""
    ARPES_Motors.move('phi',val,**kwargs)

def dmvphi(val):
    """ relative phi motor in the endstation"""
    ARPES_Motors.move('phi',val,relative=True)

def scanphi(start,stop,step,**kwargs):
    """ scans the phi motor in the endstation"""
    ARPES_Motors.scan("phi",start,stop,step,**kwargs)

def dscanphi(start,stop,step):
    """relative scans of the phi motor in the endstation """
    ARPES_Motors.scan("phi",start,stop,step,relative=True)


def mvfocus(x_val):
    """ 
    Moves APPES x and compensates y motor so that the beam stays in the same sample position but the focus is moved
    """
    x_delta = x_val - ARPES_Motors.get('x')
    y_delta = x_delta * np.tan(55/180*np.pi)
    y_val =  ARPES_Motors.get('y') + y_delta
    
    ARPES_Motors.move("x",x_val)
    ARPES_Motors.move("y",y_val)

    return x_val,y_val

def mvy_fixed_focus(y_val):
    """ 
    Moves APPES x and compensates y motor so that the beam stays in the same sample position but the focus is moved
    """  
    y_delta = y_val - ARPES_Motors.get('y')
    x_delta = y_delta / np.tan(55/180*np.pi)
    x_val =  ARPES_Motors.get('x') + x_delta

    ARPES_Motors.move("x",x_val)
    ARPES_Motors.move("y",y_val)
    return x_val,y_val

def scanfocus(x_start,x_stop,x_step,**kwargs):
    """
    Scans APPES x and compensates y motor so that the beam stays in the same sample position
    y=x*tan(55)

    kwargs:
        execute: True/False to start the scan => True (default)
    """
    kwargs.setdefault('execute',True)

   

def ARPES_sample_map2D(step_y=0.5,step_z=0.5):
    """
    2D map of sample area in ARPES chamber

    Previously: Map_ARPES_Sample
    """
    print("Scan_ARPES_2Dmotor(\"y\",0,4,"+str(step_y)+",\"z\",12,16,"+str(step_z)+")")
    ARPES_Motors.scan2D(["y",0,4,step_y],["z",12,16,step_z])


######## mprint and mvsample ###################
def mprint():
    """
    prints current position of the physical motors
    """
    ARPES_Motors.mprint()

def mvsample(position_list):
    """
    moves the sample to the position sepcified by position_list
    position_list = ['description',x,y,z,th,chi,phi]

    Previously: sample
    """
    ARPES_Motors.mvsample(position_list)

